#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include "poly.h"
#include "stack.h"


static int verbose = 0 ;

/* Common */

typedef struct {
  stack_t *stack ;
  count_t sum ;
} common_t ;

/* Worker */
typedef struct {
  int id ;
  common_t *common ;
} worker_t ;


worker_t *alloc_worker_t(common_t *c,int id) {
  worker_t *p = malloc_check(sizeof(*p)) ;
  p->id = id ;
  p->common = c ;
  return p ;
}

void free_worker_t(worker_t *p) {
  free(p);
}

void *worker(void *p) {
  worker_t *w = (worker_t *)p ;
  common_t *c = w->common ;
  int id = w->id ;
  free_worker_t(w) ;
  void *item ;
  while ((item = pop(c->stack)) != NULL) {
    subtask_t *t = (subtask_t *)item ;
    if (verbose) {
      fprintf(stderr,"WORKER<%i> GET %i\n",id,t->id);
    }
    count_t y = run_subtask(t) ;
    free_arg(t) ;
    (void)__sync_fetch_and_add(&c->sum,y) ;
  }
  if (verbose) fprintf(stderr,"WORKER<%i> OUT\n",id) ;
  return NULL ;
}


void handle_subtask(void *env,subtask_t *z) {
  if (verbose) {
    fprintf(stderr,"PUSH SUBTASK: %i\n",z->id) ;
  }
  stack_t *stack = (stack_t *)env ;
  push(stack,copy_arg(z)) ;
}

count_t run(int n, int depth, int nprocs,int sz) {
  /* Initialise */
  common_t c ;
  c.stack = alloc_stack(sz) ;
  c.sum = 0 ;
  /* Create workers */
  for (int id = 0 ; id < nprocs ; id++) {
    worker_t *w = alloc_worker_t(&c,id) ;
    create_thread_detached(worker,w) ;
  }
  /* generate subtasks */
  int ntasks =  generate_subtasks(n, depth, c.stack, handle_subtask) ;
  if (verbose) {
    fprintf(stderr,"Generated %i subtasks\n",ntasks) ;
  }
  /* wait result */
  kill(c.stack,nprocs) ;
  /* clean */
  free_stack(c.stack) ;
  return c.sum ;
}

void usage(char *pgm)
{
  fprintf(stderr,"usage: %s [-v] [-d<int>] [-j<int>] [-s<int>] [n]\n",pgm);
  exit(-1);
}
int main(int argc, char **argv) {
  char *arg=0;
  char *pgm = argv[0];
  int n = 15 ;
  int depth = 5 ;
  int nprocs = 4 ;
  int sz = 2 ;
  while (--argc) {
    arg = *++argv;
    if (*arg == '-') {
      arg++ ;
       if (*arg == 'd') {
        if (arg[1]) {
          depth = atoi(arg+1) ;
        } else {
          usage(pgm) ;
        }
       } else if (*arg == 'p' || *arg == 'j') {
         if (arg[1]) {
           nprocs = atoi(arg+1) ;
         } else {
           usage(pgm) ;
         }
       } else if (*arg == 's') {
         if (arg[1]) {
           sz = atoi(arg+1);
         } else {
           usage(pgm) ;
         }
       } else if (*arg == 'v') {
         verbose++;
       } else {
         usage(pgm);
       }
    } else break ;      
  }
  if (argc >= 1) n = atoi(arg);
  if (n > NMAX) {
      fprintf(stderr,"argument must be <= %d\n",NMAX);
      exit(2);
  }

  count_t r = run(n,depth,nprocs,sz) ;
  printf("%" PRC "\n",r) ;
  return 0 ;
}
